import threading
import subprocess
from subprocess import PIPE
from typing import TypeVar

from amino.util.string import blue
from amino import Just, Maybe, __, _

from ribosome.trans.messages import Nop, Done, Quit, Stop, CoroutineAlg, SubProcessAsync, Fork, Coroutine
from ribosome.trans.api import trans
from ribosome.dispatch.component import Component
from ribosome.trans.message_base import pmessage, json_pmessage, Message, ToMachine
from ribosome.dispatch.transform import TransValidator

Callback = pmessage('Callback', 'func')
EnvelopeOld = pmessage('EnvelopeOld', 'message', 'to')
RunMachine = json_pmessage('RunMachine', 'machine')
KillMachine = pmessage('KillMachine', 'uuid')
RunScratchMachine = json_pmessage('RunScratchMachine', 'machine')
Init = pmessage('Init')
IfUnhandled = pmessage('IfUnhandled', 'msg', 'unhandled')
A = TypeVar('A')
D = TypeVar('D')


class InternalC(Component):

    @trans.msg.unit(Nop)
    def _nop(self):
        pass

    @trans.msg.unit(Stop)
    def _stop_msg(self):
        return Quit(), Done().pub.at(1)

    @trans.msg.unit(Done)
    def _done_msg(self):
        self._done()

    @trans.msg.unit(Callback)
    def message_callback(self):
        return self.msg.func(self.data)

    @trans.msg.unit(Coroutine)
    def _couroutine(self):
        return self.msg

    @trans.msg.unit(CoroutineAlg)
    def message_couroutine_alg(self):
        async def run_coro_alg() -> None:
            res = await self.msg.coro
            trans_desc = blue(f'{self.name}.message_couroutine_alg')
            return Just(TransValidator(trans_desc).validate(res, self.data))
        return run_coro_alg()

    @trans.msg.one(SubProcessAsync)
    def message_sub_process_async(self) -> None:
        def subproc_async() -> Message:
            job = self.msg.job
            proc = subprocess.run(
                args=job.args.cons(job.exe),
                stdin=PIPE,
                stdout=PIPE,
                stderr=PIPE,
                cwd=str(job.cwd),
                **job.kw,
            )
            self.log.debug(f'finished async subproc with {proc}')
            result = job.result_strict(proc.returncode, proc.stdout or '', proc.stderr or '')
            return self.msg.result(result)
        return Fork(subproc_async)

    @trans.msg.unit(Fork)
    def message_fork(self) -> None:
        def dispatch() -> None:
            try:
                self.msg.callback() % self.send
            except Exception as e:
                self.log.caught_exception(f'running forked function {self.msg.callback}', e)
        threading.Thread(target=dispatch).start()

    @trans.msg.unit(RunMachine)
    def message_run_machine(self):
        self.sub = self.sub.cat(self.msg.machine)
        init = self.msg.options.get('init') | Init()
        return EnvelopeOld(init, self.msg.machine.uuid)

    @trans.msg.unit(KillMachine)
    def message_kill_machine(self):
        self.sub = self.sub.filter_not(_.uuid == self.msg.uuid)

    @trans.msg.unit(EnvelopeOld)
    def message_envelope(self):
        return self.sub.find(_.uuid == self.msg.to) / __.loop_process(self.data, self.msg.message)

    @trans.msg.unit(ToMachine)
    def message_to_machine(self) -> Maybe:
        return self.sub.find(_.name == self.msg.target) / __.loop_process(self.data, self.msg.message)

    @trans.msg.unit(IfUnhandled)
    def if_unhandled(self):
        result = self._send(self.data, self.msg.msg)
        return result if result.handled else self._send(self.data, self.msg.unhandled)

__all__ = ('InternalC', 'RunMachine', 'KillMachine', 'RunScratchMachine')



from amino.dat import ADT
from amino import Boolean


class PrefixStyle(ADT['PrefixStyle']):

    @property
    def short(self) -> Boolean:
        return Boolean.isinstance(self, Short)

    @property
    def full(self) -> Boolean:
        return Boolean.isinstance(self, Full)

    @property
    def plain(self) -> Boolean:
        return Boolean.isinstance(self, Plain)


class Short(PrefixStyle): pass


class Full(PrefixStyle): pass


class Plain(PrefixStyle): pass


__all__ = ('PrefixStyle', 'Short', 'Full', 'Plain')



from typing import Generator, Union, Type

from amino import Try, _, Either, List, Left, do, Right, curried
from amino.mod import cls_from_module

from ribosome.logging import Logging, ribo_log
from ribosome.nvim import NvimIO
from ribosome import NvimFacade
from ribosome.config import Config
from ribosome.dispatch.component import Component


class ComponentResolver(Logging):

    def __init__(self, config: Config) -> None:
        self.config = config

    @property
    def name(self) -> str:
        return self.config.name

    def find_component(self, name: str) -> Either[List[str], Component]:
        mods = List(
            Either.import_module(name),
            Either.import_module(f'{self.name}.components.{name}'),
            Either.import_module(f'{self.name}.plugins.{name}'),
        )
        # TODO .traverse(_.swap).swap
        errors = mods.filter(_.is_left) / _.value
        return mods.find(_.is_right) | Left(errors)

    def inst_component(self, mod):
        return (
            Try(getattr(mod, 'Component'), self.vim, self)
            if hasattr(mod, 'Component')
            else Left('module does not define class `Component`')
        )

    def extra_component(self, name: str, vim: NvimFacade) -> Either[List[str], Component]:
        auto = f'{self.config.name}.components.{name}'
        return (
            self.declared_component(name)
            .accum_error_f(lambda: self.component_from_exports(auto).lmap(List))
            .accum_error_f(lambda: self.component_from_exports(name).lmap(List))
            .flat_map(self.inst_auto(name, vim))
        )

    def declared_component(self, name: str) -> Either[List[str], Component]:
        return (
            self.config.components
            .lift(name)
            .to_either(List(f'no auto component defined for `{name}`'))
        )

    @do(Either[str, Component])
    def component_from_exports(self, mod: str) -> Generator:
        mod = yield Either.import_module(mod)
        yield cls_from_module(mod, Component)

    @curried
    def inst_auto(self, name: str, vim: NvimFacade, plug: Union[str, Type]) -> Either[str, Component]:
        return (
            Right(plug(name))
            if isinstance(plug, type) and issubclass(plug, Component) else
            Left(List(f'invalid tpe for auto component: {plug}'))
        )

    @property
    @do(NvimIO[List[str]])
    def components(self) -> Generator:
        from_user = yield self.config.settings.components.value
        additional = from_user | self.config.default_components
        components = self.config.core_components + additional
        ribo_log.debug(f'starting {self.config} with components {components}')
        yield NvimIO.pure(components)

    def create_components(self, name: str) -> NvimIO[List[Component]]:
        def report(errs):
            msg = 'invalid {} component module "{}": {}'
            self.log.error(msg.format(self.name, name, errs))
        return NvimIO.delay(
            lambda v:
            (self.find_component(name) // self.inst_component)
            .lmap(List)
            .accum_error_f(lambda: self.extra_component(name, v))
            .leffect(report)
        )

    @property
    @do(NvimIO[List[Component]])
    def run(self) -> Generator:
        comp = yield self.components
        sub = yield comp.traverse(self.create_components, NvimIO)
        yield NvimIO.from_either(sub.sequence(Either))


__all__ = ('ComponentResolver',)



import abc
from typing import TypeVar, Union, Any, Generic, Type, Callable

from amino import List, Boolean, Nil, Maybe, Just, Nothing, IO, Either, Right, Left
from amino.dat import Dat, ADT, DatMeta

from ribosome.request.rpc import RpcHandlerFunction, RpcHandlerSpec
from ribosome.request.handler.handler import RequestHandler
from ribosome.request.handler.dispatcher import TransDispatcher
from ribosome.request.handler.method import RpcMethod
from ribosome.nvim import NvimIO
from ribosome.nvim.io import NvimIOState
from ribosome.trans.message_base import Message
from ribosome.trans.effect import GatherIOs, GatherSubprocs
from ribosome.trans.action import TransDo

B = TypeVar('B')
Meth = TypeVar('Meth', bound=RpcMethod)
D = TypeVar('D')


class Dispatch:

    @property
    def sync(self) -> Boolean:
        return Boolean.isinstance(self, DispatchSync)

    @property
    def async(self) -> Boolean:
        return Boolean.isinstance(self, DispatchAsync)

    @abc.abstractmethod
    def spec(self, name: str, prefix: str) -> RpcHandlerSpec:
        ...

    @abc.abstractproperty
    def desc(self) -> str:
        ...


DP = TypeVar('DP', bound=Dispatch)


class DispatchSync(ADT['DispatchSync'], Dispatch, base=True):
    pass


class DispatchAsync(ADT['DispatchAsync'], Dispatch, base=True):
    pass


class Legacy(DispatchSync):

    def __init__(self, handler: RpcHandlerFunction) -> None:
        self.handler = handler

    def _arg_desc(self) -> List[str]:
        return List(str(self.handler))

    @property
    def sync(self) -> Boolean:
        return self.handler.spec.sync

    def spec(self, name: str, prefix: str) -> RpcHandlerSpec:
        return self.handler.spec

    @property
    def desc(self) -> str:
        return f'legacy `{self.handler}`'


class SendMessage(DispatchAsync):

    def __init__(self, handler: RequestHandler) -> None:
        self.handler = handler

    def _arg_desc(self) -> List[str]:
        return List(self.name, str(self.handler))

    @property
    def name(self) -> str:
        return self.handler.name

    @property
    def sync(self) -> Boolean:
        return self.handler.sync

    def spec(self, name: str, prefix: str) -> RpcHandlerSpec:
        return self.handler.spec(name, prefix)

    @property
    def msg(self) -> Type[Message]:
        return self.handler.dispatcher.msg

    @property
    def method(self) -> str:
        return self.handler.method.method

    @property
    def desc(self) -> str:
        return f'send `{self.msg.__name__}`'


class Trans(Generic[Meth, B], DispatchSync, DispatchAsync):

    def __init__(self, handler: RequestHandler[Meth, TransDispatcher[B]]) -> None:
        self.handler = handler

    def _arg_desc(self) -> List[str]:
        return List(self.name, str(self.handler))

    def spec(self, name: str, prefix: str) -> RpcHandlerSpec:
        return self.handler.spec(name, prefix)

    @property
    def name(self) -> str:
        return self.handler.name

    @property
    def desc(self) -> str:
        return f'trans `{self.name}`'


class Internal(Generic[Meth, B], DispatchSync, DispatchAsync):

    def __init__(self, handler: RequestHandler[Meth, TransDispatcher[B]]) -> None:
        self.handler = handler

    def _arg_desc(self) -> List[str]:
        return List(str(self.handler))

    def spec(self, name: str, prefix: str) -> RpcHandlerSpec:
        return self.handler.spec(name, prefix)

    @property
    def name(self) -> str:
        return self.handler.name

    @property
    def desc(self) -> str:
        return f'internal `{self.name}`'


class DispatchOutput(ADT, base=True): pass


class DispatchError(DispatchOutput):

    @staticmethod
    def cons(problem: Union[str, Exception]) -> 'DispatchError':
        msg = 'fatal error while handling request'
        return (
            DispatchError(msg, Just(problem))
            if isinstance(problem, Exception) else
            DispatchError(problem, Nothing)
        )

    def __init__(self, message: str, exception: Maybe[Exception]) -> None:
        self.message = message
        self.exception = exception

    def _arg_desc(self) -> List[str]:
        return List(self.message)


class DispatchErrors(DispatchOutput):

    def __init__(self, errors: List[DispatchError]) -> None:
        self.errors = errors


class DispatchReturn(DispatchOutput):

    def __init__(self, value: Any) -> None:
        self.value = value

    def _arg_desc(self) -> List[str]:
        return List(self.value)


class DispatchUnit(DispatchOutput):

    def _arg_desc(self) -> List[str]:
        return Nil


class DispatchOutputAggregate(DispatchOutput):

    def __init__(self, results: List['DispatchResult']) -> None:
        self.results = results


class DispatchDo(DispatchOutput):

    def __init__(self, trans: TransDo) -> None:
        self.trans = trans


A = TypeVar('A')
R = TypeVar('R')
I = TypeVar('I')


class DIO(Generic[I], ADT['DIO[I]'], base=True):

    @staticmethod
    def cons(io: I) -> Either[str, 'DIO[I]']:
        return (
            Right(IODIO(io))
            if isinstance(io, IO) else
            Right(GatherIOsDIO(io))
            if isinstance(io, GatherIOs) else
            Right(GatherSubprocsDIO(io))
            if isinstance(io, GatherSubprocs) else
            Right(NvimIODIO(io))
            if isinstance(io, NvimIO) else
            Left(f'invalid type for DIO: {io}')
        )


class IODIO(Generic[A], DIO[IO[A]]):

    def __init__(self, io: IO[A]) -> None:
        self.io = io


class GatherIOsDIO(Generic[A], DIO[GatherIOs[A]]):

    def __init__(self, io: GatherIOs[A]) -> None:
        self.io = io


class GatherSubprocsDIO(Generic[A, R], DIO[GatherSubprocs[A, R]]):

    def __init__(self, io: GatherSubprocs[A, R]) -> None:
        self.io = io


class NvimIODIO(Generic[A], DIO[NvimIO[A]]):

    def __init__(self, io: NvimIO[A]) -> None:
        self.io = io


class DispatchIO(Generic[I], DispatchOutput):

    def __init__(self, io: DIO[I]) -> None:
        self.io = io


class DispatchContinuation(ADT['DispatchContinuation'], base=True):
    pass


class DispatchDone(DispatchContinuation):
    pass


class DispatchHandle(DispatchContinuation):
    pass


class DispatchResultMeta(DatMeta):

    @property
    def unit(self) -> 'DispatchResult':
        return DispatchResult(DispatchUnit(), Nil)

    @property
    def unit_nio(self) -> 'NvimIOState[D, DispatchResult]':
        return NvimIOState.pure(DispatchResult.unit)

    def error(self, problem: Union[str, Exception]) -> 'NvimIOState[D, DispatchResult]':
        return DispatchResult(DispatchError.cons(problem), Nil)

    def error_nio(self, problem: Union[str, Exception]) -> 'NvimIOState[D, DispatchResult]':
        return NvimIOState.pure(DispatchResult.error(problem))


class DispatchResult(Dat['DispatchResult'], metaclass=DispatchResultMeta):

    def __init__(self, output: DispatchOutput, msgs: List[Message]) -> None:
        self.output = output
        self.msgs = msgs


__all__ = ('DispatchSync', 'DispatchAsync', 'Legacy', 'SendMessage', 'Trans', 'Internal', 'DispatchOutput',
           'DispatchError', 'DispatchReturn', 'DispatchUnit', 'DispatchContinuation', 'DispatchResult')



import abc
from typing import Callable, Type, TypeVar, Generic, Generator

from amino import List, Either, __, Left, Eval
from amino.util.string import ToStr
from amino.do import do, Do

from ribosome.nvim import NvimIO, NvimFacade
from ribosome.logging import Logging

A = TypeVar('A', contravariant=True)
B = TypeVar('B')


class PluginSetting(Generic[B], Logging, ToStr):

    @abc.abstractproperty
    def value(self) -> NvimIO[Either[str, B]]:
        ...

    @abc.abstractproperty
    def default_e(self) -> Either[str, B]:
        ...

    @property
    def value_or_default(self) -> NvimIO[B]:
        @do(NvimIO[B])
        def run() -> Generator:
            value = yield self.value
            yield NvimIO.from_either(value.o(self.default_e))
        return run()


class StrictSetting(Generic[A, B], PluginSetting[B]):

    def __init__(
            self,
            name: str,
            desc: str,
            help: str,
            prefix: bool,
            tpe: Type[A],
            ctor: Callable[[A], B],
            default: Either[str, B],
    ) -> None:
        self.name = name
        self.desc = desc
        self.help = help
        self.prefix = prefix
        self.tpe = tpe
        self.ctor = ctor
        self.default = default

    @property
    def value(self) -> NvimIO[Either[str, B]]:
        @do(Either[str, B])
        def read(v: NvimFacade) -> Do:
            vars = v.vars
            getter = vars.p if self.prefix else vars
            raw = yield vars.typed(self.tpe, getter(self.name))
            yield self.ctor(raw)
        return NvimIO.delay(read)

    def value_or(self, default: B) -> NvimIO[B]:
        return self.value / __.get_or_else(default)

    @property
    def default_e(self) -> Either[str, B]:
        return self.default

    def _arg_desc(self) -> List[str]:
        return List(self.name, str(self.prefix), str(self.tpe))


class EvalSetting(Generic[B], PluginSetting[B]):

    def __init__(
            self,
            name: str,
            f: Eval[NvimIO[Either[str, B]]],
            default: Either[str, B]=Left('no default specified')
    ) -> None:
        self.name = name
        self.f = f
        self.default = default

    def _arg_desc(self) -> List[str]:
        return List(self.name)

    @property
    def value(self) -> NvimIO[Either[str, B]]:
        return self.f.value

    @property
    def default_e(self) -> Either[str, B]:
        return self.default


def setting_ctor(tpe: Type[A], ctor: Callable[[A], B]) -> Callable[[str, str, str, bool, B], PluginSetting[B]]:
    def setting(name: str, desc: str, help: str, prefix: bool, default: Either[str, B]=Left('no default specified')
                ) -> PluginSetting[B]:
        return StrictSetting(name, desc, help, prefix, tpe, ctor, default)
    return setting


__all__ = ('PluginSetting', 'StrictSetting', 'EvalSetting', 'setting_ctor')



from typing import Callable, TypeVar, Type, Generic, Any

from amino import List, Lists
from amino.dat import Dat

from ribosome.trans.effect import TransEffect, cont, lift
from ribosome.trans.action import TransAction, TransM, TransMPure
from ribosome.trans.message_base import Message
from ribosome.dispatch.component import Component

D = TypeVar('D')
M = TypeVar('M', bound=Message)
R = TypeVar('R')
O = TypeVar('O')
A = TypeVar('A')
B = TypeVar('B')


class TransComplete(Dat['TransComplete']):

    def __init__(self, name: str, action: TransAction) -> None:
        self.name = name
        self.action = action


def extract(name: str, output: O, effects: List[TransEffect]) -> TransComplete:
    trans_result = cont(effects, False, lambda f: f(output)) | output
    return TransComplete(name, lift(trans_result, False))


class TransHandler:
    pass


class MessageTransHandler(Generic[M, D], Dat['MessageTransHandler[M, D]'], TransHandler):

    @staticmethod
    def create(fun: Callable[[M], R], msg: Type[M], effects: List[TransEffect], prio: float) -> 'TransHandler':
        name = fun.__name__
        return MessageTransHandler(name, fun, msg, effects, prio)

    def __init__(self, name: str, fun: Callable[[M], R], message: Type[M], effects: List[TransEffect], prio: float
                 ) -> None:
        self.name = name
        self.message = message
        self.fun = fun
        self.prio = prio
        self.effects = effects

    def run(self, component: Component, msg: M) -> TransAction:
        return extract(self.name, self.fun(component, msg), Lists.wrap(self.effects))


class FreeTransHandler(Generic[D, R], Dat['FreeTransHandler[M, D]'], TransHandler):

    @staticmethod
    def create(fun: Callable[..., R], effects: List[TransEffect], prio: float) -> 'TransHandler':
        name = fun.__name__
        return FreeTransHandler(name, fun, (), effects, prio)

    def __init__(self, name: str, fun: Callable[..., R], args: tuple, effects: List[TransEffect], prio: float) -> None:
        self.name = name
        self.fun = fun
        self.args = args
        self.effects = effects
        self.prio = prio

    def __call__(self, *args: Any) -> 'FreeTransHandler':
        return self.copy(args=args)

    def run(self) -> TransAction:
        return extract(self.name, self.fun(*self.args), Lists.wrap(self.effects))

    @property
    def m(self) -> TransM:
        return TransMPure(self)


__all__ = ('MessageTransHandler', 'FreeTransHandler')



import abc

from amino import List, __, _

from ribosome.unite import UniteMessage, UniteSource
from ribosome.unite.data import UniteKind, UniteSyntax
from ribosome.trans.api import trans
from ribosome.trans.messages import UnitIO
from ribosome.dispatch.component import Component


class Unite(Component):

    @trans.msg.one(UniteSyntax, trans.m)
    def syntax(self):
        return self.machine.syntax(self.msg.source) / UnitIO


class UniteMachine:

    def __init__(self) -> None:
        self._unite_ready = False

    def prepare(self, msg):
        if not self._unite_ready and isinstance(msg, UniteMessage):
            self._setup_unite()

    def _setup_unite(self):
        self.sources % __.define(self.vim)
        self.kinds % __.define(self.vim)
        self._unite_ready = True

    @abc.abstractproperty
    def sources(self) -> List[UniteSource]:
        ...

    @abc.abstractproperty
    def kinds(self) -> List[UniteKind]:
        ...

    def syntax(self, source):
        return (self.sources.find(_.name == source) /
                __.syntax_task(self.vim.buffer.syntax))

__all__ = ('Unite',)



import abc
from typing import Type, TypeVar

from amino.dat import ADT

from ribosome.request.rpc import RpcHandlerSpec, RpcCommandSpec, RpcFunctionSpec, RpcAutocommandSpec

RHS = TypeVar('RHS', bound=RpcHandlerSpec)


class RpcMethod(ADT['RpcMethod'], base=True):

    @abc.abstractproperty
    def spec_type(self) -> Type[RHS]:
        ...

    @abc.abstractproperty
    def method(self) -> str:
        ...


class CmdMethod(RpcMethod):

    @property
    def spec_type(self) -> Type[RHS]:
        return RpcCommandSpec

    @property
    def method(self) -> str:
        return 'command'


class FunctionMethod(RpcMethod):

    @property
    def spec_type(self) -> Type[RHS]:
        return RpcFunctionSpec

    @property
    def method(self) -> str:
        return 'function'


class AutocmdMethod(RpcMethod):

    @property
    def spec_type(self) -> Type[RHS]:
        return RpcAutocommandSpec

    @property
    def method(self) -> str:
        return 'autocmd'


__all__ = ('RpcMethod', 'CmdMethod', 'FunctionMethod', 'AutocmdMethod')



from amino import Right, List, __, Map

from ribosome.test.integration.klk import later

from kallikrein import kf, k, Expectation
from kallikrein.matchers.either import be_right, be_left
from kallikrein.matchers.maybe import be_just

from integration._support.facade import FacadeTestPlugin
from integration._support.base import IntegrationSpecBase


class VimSpec(IntegrationSpecBase):
    '''NvimFacade
    get and set variables $vars
    '''

    @property
    def _prefix(self):
        return 'ribosome'

    @property
    def plugin_class(self):
        return Right(FacadeTestPlugin)

    def vars(self) -> Expectation:
        vname = List.random_string()
        content = [List.random_string()]
        self.vim.cmd_once_defined('Go')
        self.vim.buffer.vars.set_p(vname, content)
        later(kf(self.vim.buffer.vars.pl, vname).must(be_right(content)))
        var = (
            self.vim.call('AllVars') /
            Map //
            __.get('b:{}_{}'.format(self._prefix, vname))
        )
        return (
            k(var).must(be_just(str(content))) &
            kf(self.vim.vars.p, vname).must(be_left)
        )

__all__ = ('VimSpec',)



from typing import Callable, Tuple, Any, TypeVar

from ribosome.nvim import NvimIO
from ribosome.plugin_state import PluginState, PluginStateHolder

from toolz import assoc

from amino import Lists, Left, Map
from amino.boolean import true
from amino.dat import Dat
from ribosome.dispatch.run import DispatchJob
from ribosome.dispatch.data import Dispatch, DispatchResult, DIO
from ribosome.test.spec import MockNvimFacade
from ribosome.host import init_state, dispatch_job, HostConfig, host_config
from ribosome.dispatch.execute import sync_runner, async_runner, sync_sender, run_dispatch, Res, execute_async_loop
from ribosome.nvim.io import NvimIOState, NS
from ribosome.config import Config
from ribosome.trans.handler import TransComplete

D = TypeVar('D')


class DispatchHelper(Dat['DispatchHelper']):

    @staticmethod
    def cons(
            config: Config,
            *comps: str,
            vars: dict=dict(),
            responses: Map[str, Any]=Map(),
            io_executor: Callable[[DIO], NS[PluginState[D], TransComplete]]=None,
    ) -> 'DispatchHelper':
        hc = host_config(config, Left('spec'), true, io_executor)
        comps_var = (f'{config.name}_components', Lists.wrap(comps))
        vim = MockNvimFacade(prefix=config.name, vars=assoc(vars, *comps_var), responses=responses)
        state = init_state(hc).unsafe(vim)
        return DispatchHelper(vim, hc, state)

    def __init__(self, vim: MockNvimFacade, host_conf: HostConfig, state: PluginState) -> None:
        self.vim = vim
        self.host_conf = host_conf
        self.state = state

    def dispatch(self, name: str, args: tuple) -> Tuple[DispatchJob, Dispatch]:
        holder = PluginStateHolder.cons(self.state)
        job = dispatch_job(True, self.host_conf.async_dispatch, holder, name, self.state.config.prefix, (args,))
        return job, job.dispatches.lift(job.name).get_or_fail('no matching dispatch')

    def sender(self, name: str, args: tuple=(), sync: bool=True) -> Callable[[], Res]:
        job, dispatch = self.dispatch(name, args)
        runner = sync_runner if sync else async_runner
        return sync_sender(job, dispatch, runner)

    def run(self, name: str, args: tuple=(), sync: bool=True) -> NvimIOState[PluginState, DispatchResult]:
        send = self.sender(name, args, sync)
        return run_dispatch(send, NvimIOState.pure)

    def loop(self, name: str, args: tuple=(), sync: bool=True) -> NvimIO[Tuple[PluginState, DispatchResult]]:
        job, dispatch = self.dispatch(name, args)
        return execute_async_loop(job, dispatch)

    def unsafe_run(self, name: str, args=(), sync: bool=True) -> Tuple[PluginState, DispatchResult]:
        return self.run(name, args=args, sync=sync).run(self.state).unsafe(self.vim)


__all__ = ('DispatchHelper',)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# __coconut_hash__ = 0xf5f7bde2

# Compiled with Coconut version 1.3.0 [Dead Parrot]

# Coconut Header: -------------------------------------------------------------

from __future__ import generator_stop
import sys as _coconut_sys, os.path as _coconut_os_path
_coconut_file_path = _coconut_os_path.dirname(_coconut_os_path.abspath(__file__))
_coconut_sys.path.insert(0, _coconut_file_path)
from __coconut__ import _coconut, _coconut_NamedTuple, _coconut_MatchError, _coconut_tail_call, _coconut_tco, _coconut_igetitem, _coconut_base_compose, _coconut_forward_compose, _coconut_back_compose, _coconut_forward_star_compose, _coconut_back_star_compose, _coconut_pipe, _coconut_star_pipe, _coconut_back_pipe, _coconut_back_star_pipe, _coconut_bool_and, _coconut_bool_or, _coconut_none_coalesce, _coconut_minus, _coconut_map, _coconut_partial
from __coconut__ import *
_coconut_sys.path.remove(_coconut_file_path)

# Compiled Coconut: -----------------------------------------------------------

from typing import TypeVar

from amino import Just
from amino import Nothing
from amino import Right
from amino import Left
from amino import L
from amino import _
from amino import __
from amino import Nil
from amino.state import EvalState
from amino.state import EvalState
from amino.do import do

from ribosome.logging import ribo_log
from ribosome.plugin_state import ComponentState
from ribosome.plugin_state import PluginState
from ribosome.plugin_state import TransState
from ribosome.nvim.io import NvimIOState
from ribosome.dispatch.data import DispatchResult
from ribosome.dispatch.data import DispatchUnit
from ribosome.trans.message_base import Message
from ribosome.trans.queue import PrioQueue

A = TypeVar('A')
D = TypeVar('D')


@_coconut_tco
def process_message(messages: 'PrioQueue[A]', send: '_coconut.typing.Callable[[A], NvimIOState[D, DispatchResult]]') -> '(PrioQueue[A], NvimIOState[D, DispatchResult])':
    def process1(*_coconut_match_to_args, **_coconut_match_to_kwargs):
        _coconut_match_check = False
        if (_coconut.len(_coconut_match_to_args) == 1) and (_coconut.isinstance(_coconut_match_to_args[0], Right)) and (_coconut.len(_coconut_match_to_args[0]) == 1) and (_coconut.isinstance(_coconut_match_to_args[0][0], _coconut.abc.Sequence)) and (_coconut.len(_coconut_match_to_args[0][0]) == 2) and (_coconut.isinstance(_coconut_match_to_args[0][0][0], _coconut.abc.Sequence)) and (_coconut.len(_coconut_match_to_args[0][0][0]) == 2):
            prio = _coconut_match_to_args[0][0][0][0]
            msg = _coconut_match_to_args[0][0][0][1]
            rest = _coconut_match_to_args[0][0][1]
            if not _coconut_match_to_kwargs:
                _coconut_match_check = True
        if not _coconut_match_check:
            _coconut_match_err = _coconut_MatchError("pattern-matching failed for " "'def process1(Right(((prio, msg), rest))) = rest, send(msg)'" " in " + _coconut.repr(_coconut.repr(_coconut_match_to_args)))
            _coconut_match_err.pattern = 'def process1(Right(((prio, msg), rest))) = rest, send(msg)'
            _coconut_match_err.value = _coconut_match_to_args
            raise _coconut_match_err

        return rest, send(msg)
    @addpattern(process1)
    def process1(*_coconut_match_to_args, **_coconut_match_to_kwargs):
        _coconut_match_check = False
        if (_coconut.len(_coconut_match_to_args) == 1) and (_coconut.isinstance(_coconut_match_to_args[0], Left)) and (_coconut.len(_coconut_match_to_args[0]) == 1):
            err = _coconut_match_to_args[0][0]
            if not _coconut_match_to_kwargs:
                _coconut_match_check = True
        if not _coconut_match_check:
            _coconut_match_err = _coconut_MatchError("pattern-matching failed for " "'def process1(Left(err)) = messages, NvimIOState.pure(DispatchResult(DispatchUnit(), Nil))'" " in " + _coconut.repr(_coconut.repr(_coconut_match_to_args)))
            _coconut_match_err.pattern = 'def process1(Left(err)) = messages, NvimIOState.pure(DispatchResult(DispatchUnit(), Nil))'
            _coconut_match_err.value = _coconut_match_to_args
            raise _coconut_match_err

        return messages, NvimIOState.pure(DispatchResult(DispatchUnit(), Nil))
    return _coconut_tail_call(process1, messages.get)
